@@ -1,4 +1,5 @@
 import os
+import random
 from os.path import basename, dirname
 import time
 import logging
@@ -36,7 +37,7 @@ def tagging_metrics(task_name, preds, labels):
     acc = num_correct / len(preds)
     f1 = f1_score(labels, preds, average=None)
 
-    return {'acc': acc, 'token_f1': f1, 'f1': seq_f1([pred_seq], [label_seq]), 'report':'\n'+seq_cls([pred_seq], [label_seq])}
+    return {'acc': acc, 'token_f1': fix_np_types(f1), 'f1': fix_np_types(seq_f1([label_seq], [pred_seq])), 'report':'\n'+seq_cls([label_seq], [pred_seq])}
 
 def relation_metrics(task_name, preds, labels):
 
@@ -53,11 +54,11 @@ def relation_metrics(task_name, preds, labels):
     num_correct = (relevant_labels == relevant_preds).sum()
     acc = num_correct / len(relevant_preds)
 
-    recall = recall_score(relevant_preds, relevant_labels, average=None)
-    precision = precision_score(relevant_preds, relevant_labels, average=None)
-    f1_report = f1_score(relevant_labels, relevant_preds, average=None)
+    recall = recall_score(y_pred=relevant_preds, y_true=relevant_labels, average=None)
+    precision = precision_score(y_pred=relevant_preds, y_true=relevant_labels, average=None)
+    f1_report = f1_score(y_true=relevant_labels, y_pred=relevant_preds, average=None)
 
-    return {'f1': f1_report, 'acc': acc, 'recall':recall, 'precision':precision }
+    return {'f1': fix_np_types(f1_report), 'acc': acc, 'recall':fix_np_types(recall), 'precision':fix_np_types(precision) }
 
 def fix_np_types(input_variable):
     ''' in the mtl classification setting, f1 is an array, and when the HF library
@@ -104,7 +105,7 @@ def cnlp_compute_metrics(task_name, preds, labels):
     elif task_name == 'timecat':
         return acc_and_f1(preds, labels)
     elif task_name.startswith('i2b22008'):
-        return acc_and_f1(preds, labels)
+        return { 'f1': fix_np_types(f1_score(y_true=labels, y_pred=preds, average=None))} #acc_and_f1(preds, labels)
     elif task_name == 'timex' or task_name == 'event' or task_name == 'dphe':
         return tagging_metrics(task_name, preds, labels)
     elif task_name == 'tlink-sent':
@@ -384,7 +385,7 @@ class i2b22008Processor(MTLClassifierProcessor):
         return len(self.get_classifiers())
     
     def get_one_score(self, results):
-        return results['f1'].mean()
+        return np.mean(results['f1'])
 
 cnlp_processors = {'polarity': NegationProcessor,
                    'uncertainty': UncertaintyProcessor,
