@@ -35,7 +35,7 @@ from seqeval.metrics.sequence_labeling import get_entities
 
 import torch
 from torch.utils.data.dataset import Dataset
-from transformers import AutoConfig, AutoModelForSequenceClassification, AutoTokenizer, EvalPrediction
+from transformers import AutoConfig, AutoTokenizer, AutoModel, EvalPrediction
 from transformers.training_args import IntervalStrategy
 from transformers.data.processors.utils import InputFeatures
 from transformers.tokenization_utils import PreTrainedTokenizer
@@ -44,12 +44,16 @@ from transformers.data.processors.utils import DataProcessor, InputExample, Inpu
 from transformers import ALL_PRETRAINED_CONFIG_ARCHIVE_MAP
 from transformers.optimization import AdamW, get_scheduler
 from transformers.trainer_pt_utils import get_parameter_names
+from transformers.file_utils import hf_bucket_url, CONFIG_NAME
 
 from .cnlp_processors import cnlp_processors, cnlp_output_modes, cnlp_compute_metrics, tagging, relex, classification
 from .cnlp_data import ClinicalNlpDataset, DataTrainingArguments
 
-from .CnlpRobertaForClassification import CnlpRobertaForClassification
+from .CnlpModelForClassification import CnlpModelForClassification, CnlpConfig
 from .BaselineModels import CnnSentenceClassifier, LstmSentenceClassifier
+from .HierarchicalTransformer import HierarchicalModel, HierarchicalTransformerConfig
+
+import requests
 
 from transformers import (
     HfArgumentParser,
@@ -58,6 +62,7 @@ from transformers import (
     set_seed,
 )
 
+cnlpt_models = ['cnn', 'lstm', 'hier', 'cnlpt']
 
 logger = logging.getLogger(__name__)
 
@@ -72,14 +77,25 @@ class CnlpTrainingArguments(TrainingArguments):
     final_task_weight: Optional[float] = field(
         default=1.0, metadata={"help": "Amount to up/down-weight final task in task list (other tasks weighted 1.0)"}
     )
+    freeze: bool = field(
+        default=False, metadata={"help": "Freeze the encoder layers and only train the layer between the encoder and classification architecture. Probably works best with --token flag since [CLS] may not be well-trained for anything in particular."}
+    )
+    arg_reg: Optional[float] = field(
+        default=-1, metadata={"help": "Weight to use on argument regularization term (penalizes end-to-end system if a discovered relation has low probability of being any entity type). Value < 0 (default) turns off this penalty."}
+    )
+    bias_fit: bool = field(
+        default=False, metadata={"help": "Only optimize the bias parameters of the encoder (and the weights of the classifier heads), as proposed in the BitFit paper by Ben Zaken et al. 2021 (https://arxiv.org/abs/2106.10199)"}
+    )
 
 @dataclass
 class ModelArguments:
     """
     Arguments pertaining to which model/config/tokenizer we are going to fine-tune from.
     """
-
-    model_name_or_path: str = field(
+    model: Optional[str] = field( default='cnlpt', 
+        metadata={'help': "Model type", 'choices':cnlpt_models}
+    )
+    encoder_name: Optional[str] = field(default='roberta-base',
         metadata={"help": "Path to pretrained model or model identifier from huggingface.co/models"}
     )
     config_name: Optional[str] = field(
@@ -97,9 +113,6 @@ class ModelArguments:
     token: bool = field(
         default=False, metadata={"help": "Classify over an actual token rather than the [CLS] ('<s>') token -- requires that the tokens to be classified are surrounded by <e>/</e> tokens"}
     )
-    freeze: bool = field(
-        default=False, metadata={"help": "Freeze the encoder layers and only train the layer between the encoder and classification architecture. Probably works best with --token flag since [CLS] may not be well-trained for anything in particular."}
-    )
     num_rel_feats: Optional[int] = field(
         default=12, metadata={"help": "Number of features/attention heads to use in the NxN relation classifier"}
     )
@@ -109,9 +122,65 @@ class ModelArguments:
     use_prior_tasks: bool = field(
         default=False, metadata={"help": "In the multi-task setting, incorporate the logits from the previous tasks into subsequent representation layers. This will be done in the task order specified in the command line."}
     )
-    arg_reg: Optional[float] = field(
-        default=-1, metadata={"help": "Weight to use on argument regularization term (penalizes end-to-end system if a discovered relation has low probability of being any entity type). Value < 0 (default) turns off this penalty."}
+    hier_num_layers: Optional[int] = field(
+        default=2,
+        metadata={
+            "help": (
+                "For the hierarchical model, the number of document-level transformer "
+                "layers"
+            )
+        },
     )
+    hier_hidden_dim: Optional[int] = field(
+        default=2048,
+        metadata={
+            "help": (
+                "For the hierarchical model, the inner hidden size of the positionwise "
+                "FFN in the document-level transformer layers"
+            )
+        },
+    )
+    hier_n_head: Optional[int] = field(
+        default=8,
+        metadata={
+            "help": (
+                "For the hierarchical model, the number of attention heads in the "
+                "document-level transformer layers"
+            )
+        },
+    )
+    hier_d_k: Optional[int] = field(
+        default=8,
+        metadata={
+            "help": (
+                "For the hierarchical model, the size of the query and key vectors in "
+                "the document-level transformer layers"
+            )
+        },
+    )
+    hier_d_v: Optional[int] = field(
+        default=96,
+        metadata={
+            "help": (
+                "For the hierarchical model, the size of the value vectors in the "
+                "document-level transformer layers"
+            )
+        },
+    )
+
+
+def is_pretrained_model(model_name):
+    # check if it's a built-in pre-trained config:
+    if model_name in ALL_PRETRAINED_CONFIG_ARCHIVE_MAP:
+        return True
+    
+    # check if it's a model on the huggingface model hub:
+    url = hf_bucket_url(model_name, CONFIG_NAME)
+    r = requests.head(url)
+    if r.status_code == 200:
+        return True
+
+    return False
 
 def main():
     # See all possible arguments in src/transformers/training_args.py
@@ -138,7 +207,6 @@ def main():
 
     assert len(data_args.task_name) == len(data_args.data_dir), 'Number of tasks and data directories should be the same!'
 
-    baselines = ['cnn', 'lstm']
     
     # Setup logging
     logging.basicConfig(
@@ -188,23 +256,26 @@ def main():
 
     # Load tokenizer: Need this first for loading the datasets
     tokenizer = AutoTokenizer.from_pretrained(
-        model_args.tokenizer_name if model_args.tokenizer_name else model_args.model_name_or_path,
+        model_args.tokenizer_name if model_args.tokenizer_name else model_args.encoder_name,
         cache_dir=model_args.cache_dir,
         add_prefix_space=True,
         additional_special_tokens=['<e>', '</e>', '<a1>', '</a1>', '<a2>', '</a2>', '<cr>', '<neg>']
     )
-    
+
+    model_name = model_args.model
+    hierarchical = model_name == 'hier'
+
     # Get datasets
     train_dataset = (
-        ClinicalNlpDataset(data_args, tokenizer=tokenizer, cache_dir=model_args.cache_dir) if training_args.do_train else None
+        ClinicalNlpDataset(data_args, tokenizer=tokenizer, cache_dir=model_args.cache_dir, hierarchical=hierarchical) if training_args.do_train else None
     )
     eval_dataset = (
-        ClinicalNlpDataset(data_args, tokenizer=tokenizer, mode="dev", cache_dir=model_args.cache_dir)
+        ClinicalNlpDataset(data_args, tokenizer=tokenizer, mode="dev", cache_dir=model_args.cache_dir, hierarchical=hierarchical)
         if training_args.do_eval
         else None
     )
     test_dataset = (
-        ClinicalNlpDataset(data_args, tokenizer=tokenizer, mode="test", cache_dir=model_args.cache_dir)
+        ClinicalNlpDataset(data_args, tokenizer=tokenizer, mode="test", cache_dir=model_args.cache_dir, hierarchical=hierarchical)
         if training_args.do_predict
         else None
     )
@@ -215,71 +286,145 @@ def main():
     # The .from_pretrained methods guarantee that only one local process can concurrently
     # download model & vocab.
 
-    model_name = model_args.model_name_or_path
-    if not model_name in ALL_PRETRAINED_CONFIG_ARCHIVE_MAP and not model_name in baselines:
-        # we are loading one of our own trained models, load it as-is initially,
-        # then delete its classifier head, save as temp file, and make that temp file
-        # the model file to be loaded down below the normal way. since that temp file
-        # doesn't have a stored classifier it will use the randomly-inited classifier head
-        # with the size of the supplied config (for the new task)
-        config = AutoConfig.from_pretrained(
-            model_args.config_name if model_args.config_name else model_args.model_name_or_path,
-            cache_dir=model_args.cache_dir,
-        )
-        model = CnlpRobertaForClassification.from_pretrained(
-                model_args.model_name_or_path,
-                config=config,
-                cache_dir=model_args.cache_dir,
-                layer=model_args.layer,
-                tokens=model_args.token,
-                freeze=model_args.freeze,
-                tagger=tagger,
-                relations=relations,
-                num_attention_heads=model_args.num_rel_feats,
-                head_size=model_args.head_features,
-                class_weights=None if train_dataset is None else train_dataset.class_weights,
-                final_task_weight=training_args.final_task_weight,
-                use_prior_tasks=model_args.use_prior_tasks,
-                argument_regularization=model_args.arg_reg)
-        delattr(model, 'classifiers')
-        delattr(model, 'feature_extractors')
-        if training_args.do_train:
-            tempmodel = tempfile.NamedTemporaryFile(dir=model_args.cache_dir)
-            torch.save(model.state_dict(), tempmodel)
-            model_name = tempmodel.name
+    pretrained = False
 
+    if model_name == 'cnn':
+        model = CnnSentenceClassifier(len(tokenizer), num_labels_list=num_labels)
+    elif model_name == 'lstm':
+        model = LstmSentenceClassifier(len(tokenizer), num_labels_list=num_labels)
+    elif model_name == 'hier':
+        # encoder_config = AutoConfig.from_pretrained(
+        #     model_args.config_name if model_args.config_name else model_args.encoder_name,
+        #     finetuning_task=data_args.task_name,
+        # )
 
-    if model_name in baselines:
-        if model_name == 'cnn':
-            model = CnnSentenceClassifier(len(tokenizer), num_labels_list=num_labels)
-        elif model_name == 'lstm':
-            model = LstmSentenceClassifier(len(tokenizer), num_labels_list=num_labels)
-        
-        pretrained = False
-    else:
-        config = AutoConfig.from_pretrained(
-            model_args.config_name if model_args.config_name else model_args.model_name_or_path,
-            finetuning_task=data_args.task_name,
-        )
         pretrained = True
-        model = CnlpRobertaForClassification.from_pretrained(
-            model_name,
-            config=config,
-            num_labels_list=num_labels,
-            cache_dir=model_args.cache_dir,
+
+        encoder_name = model_args.config_name if model_args.config_name else model_args.encoder_name
+        config = CnlpConfig(
+            encoder_name,
+            data_args.task_name,
+            num_labels,
             layer=model_args.layer,
             tokens=model_args.token,
-            freeze=model_args.freeze,
+            num_rel_attention_heads=model_args.num_rel_feats,
+            rel_attention_head_dims=model_args.head_features,
             tagger=tagger,
             relations=relations,
-            num_attention_heads=model_args.num_rel_feats,
-            head_size=model_args.head_features,
+        )
+        # num_tokens=len(tokenizer))
+        config.vocab_size = len(tokenizer)
+
+        encoder_dim = config.hidden_size
+
+        transformer_head_config = HierarchicalTransformerConfig(
+            n_layers=model_args.hier_num_layers,
+            d_model=encoder_dim,
+            d_inner=model_args.hier_hidden_dim,
+            n_head=model_args.hier_n_head,
+            d_k=model_args.hier_d_k,
+            d_v=model_args.hier_d_v,
+        )
+
+        model = HierarchicalModel(
+            config=config,
+            transformer_head_config=transformer_head_config,
             class_weights=None if train_dataset is None else train_dataset.class_weights,
             final_task_weight=training_args.final_task_weight,
-            use_prior_tasks=model_args.use_prior_tasks,
-            argument_regularization=model_args.arg_reg)
-        
-        model.resize_token_embeddings(len(tokenizer))
+            freeze=training_args.freeze,
+            argument_regularization=training_args.arg_reg,
+        )
+
+    else:
+        # by default cnlpt model, but need to check which encoder they want
+        encoder_name = model_args.encoder_name
+
+        # TODO check when download any pretrained language model to local disk, if 
+        # the following condition "is_pretrained_model(encoder_name)" works or not.
+        if not is_pretrained_model(encoder_name):
+            # we are loading one of our own trained models as a starting point.
+            #
+            # 1) if training_args.do_train is true:
+            # sometimes we may want to use an encoder that has been had continued pre-training, either on
+            # in-domain MLM or another task we think might be useful. In that case our encoder will just
+            # be a link to a directory. If the encoder-name is not recognized as a pre-trianed model, special
+            # logic for ad hoc encoders follows:
+            # we will load it as-is initially, then delete its classifier head, save the encoder
+            # as a temp file, and make that temp file
+            # the model file to be loaded down below the normal way. since that temp file
+            # doesn't have a stored classifier it will use the randomly-inited classifier head
+            # with the size of the supplied config (for the new task).
+            # TODO This setting 1) is not tested yet.
+            # 2) if training_args.do_train is false:
+            # we evaluate or make predictions of our trained models. 
+            # Both two setting require the registeration of CnlpConfig, and use 
+            # AutoConfig.from_pretrained() to load the configuration file
+            AutoConfig.register("cnlpt", CnlpConfig)
+            AutoModel.register(CnlpConfig, CnlpModelForClassification)
+
+            
+            # Load the cnlp configuration using AutoConfig, this will not override 
+            # the arguments from trained cnlp models. While using CnlpConfig will override
+            # the model_type and model_name of the encoder.
+            config = AutoConfig.from_pretrained(
+                model_args.config_name if model_args.config_name else model_args.encoder_name,
+                cache_dir=model_args.cache_dir,
+            )
+
+            if training_args.do_train:
+                # Setting 1) only load weights from the encoder
+                raise NotImplementedError('This functionality has not been restored yet')
+                model = CnlpModelForClassification(
+                        model_path = model_args.encoder_name,
+                        config=config,
+                        cache_dir=model_args.cache_dir,
+                        tagger=tagger,
+                        relations=relations,
+                        class_weights=None if train_dataset is None else train_dataset.class_weights,
+                        final_task_weight=training_args.final_task_weight,
+                        use_prior_tasks=model_args.use_prior_tasks,
+                        argument_regularization=model_args.arg_reg)
+                delattr(model, 'classifiers')
+                delattr(model, 'feature_extractors')
+                if training_args.do_train:
+                    tempmodel = tempfile.NamedTemporaryFile(dir=model_args.cache_dir)
+                    torch.save(model.state_dict(), tempmodel)
+                    model_name = tempmodel.name
+            else:
+                # setting 2) evaluate or make predictions
+                model = CnlpModelForClassification.from_pretrained(
+                    model_args.encoder_name,
+                    config=config,
+                    class_weights=None if train_dataset is None else train_dataset.class_weights,
+                    final_task_weight=training_args.final_task_weight,
+                    freeze=training_args.freeze,
+                    bias_fit=training_args.bias_fit,
+                    argument_regularization=training_args.arg_reg)
+
+        else:
+            # This only works when model_args.encoder_name is one of the 
+            # model card from https://huggingface.co/models
+            # By default, we use model card as the starting point to fine-tune
+            encoder_name = model_args.config_name if model_args.config_name else model_args.encoder_name
+            config = CnlpConfig(encoder_name,
+                                data_args.task_name,
+                                num_labels,
+                                layer=model_args.layer,
+                                tokens=model_args.token,
+                                num_rel_attention_heads=model_args.num_rel_feats,
+                                rel_attention_head_dims=model_args.head_features,
+                                tagger=tagger,
+                                relations=relations,)
+                                #num_tokens=len(tokenizer))
+            config.vocab_size = len(tokenizer)
+            pretrained = True
+            model = CnlpModelForClassification(
+                config=config,
+                class_weights=None if train_dataset is None else train_dataset.class_weights,
+                final_task_weight=training_args.final_task_weight,
+                freeze=training_args.freeze,
+                bias_fit=training_args.bias_fit,
+                argument_regularization=training_args.arg_reg)
 
     best_eval_results = None
     output_eval_file = os.path.join(
@@ -299,6 +444,8 @@ def main():
             steps_per_epoch = int(total_steps // training_args.num_train_epochs)
             training_args.eval_steps = steps_per_epoch // training_args.evals_per_epoch
             training_args.evaluation_strategy = IntervalStrategy.STEPS
+            # This will save model per epoch
+            # training_args.save_strategy = IntervalStrategy.EPOCH
         elif training_args.do_eval:
             logger.info('Evaluation strategy not specified so evaluating every epoch')
             training_args.evaluation_strategy = IntervalStrategy.EPOCH
@@ -370,9 +517,6 @@ def main():
         compute_metrics=build_compute_metrics_fn(task_names, model),
     )
 
-    # from watchpoints import watch
-    # watch(trainer.control.should_save)
-
     # Training
     if training_args.do_train:
         trainer.train(
@@ -409,7 +553,7 @@ def main():
                 predictions = trainer.predict(test_dataset=eval_dataset).predictions
                 dataset_labels = eval_dataset.get_labels()
                 for task_ind, task_name in enumerate(task_names):
-                    if output_mode[task_ind] == 'classification':
+                    if output_mode[task_ind] == classification:
                         task_predictions = np.argmax(predictions[task_ind], axis=1)
                         for index, item in enumerate(task_predictions):
                             if len(task_names) > len(dataset_labels):
@@ -418,7 +562,7 @@ def main():
                                 subtask_ind = task_ind
                             item = dataset_labels[subtask_ind][item]
                             writer.write("Task %d (%s) - Index %d - %s\n" % (task_ind, task_name, index, item))
-                    elif output_mode[task_ind] == 'tagging':
+                    elif output_mode[task_ind] == tagging:
                         task_predictions = np.argmax(predictions[task_ind], axis=2)
                         task_labels = dataset_labels[task_ind]
                         for index, pred_seq in enumerate(task_predictions):
@@ -435,7 +579,19 @@ def main():
 
                             entities = get_entities(chunk_labels)
                             writer.write('Task %d (%s) - Index %d: %s\n' % (task_ind, task_name, index, str(entities)))
-
+                    elif output_mode[task_ind] == relex:
+                        task_predictions = np.argmax(predictions[task_ind], axis=3)
+                        task_labels = dataset_labels[task_ind]
+                        assert task_labels[0] == 'None', 'The first labeled relation category should always be "None" but for task %s it is %s' % (task_names[task_ind], task_labels[0])
+                        
+                        for inst_ind in range(task_predictions.shape[0]):
+                            inst_preds = task_predictions[inst_ind]
+                            a1s, a2s = np.where(inst_preds > 0)
+                            for arg_ind in range(len(a1s)):
+                                a1_ind = a1s[arg_ind]
+                                a2_ind = a2s[arg_ind]
+                                cat = task_labels[ inst_preds[a1_ind][a2_ind] ]
+                                writer.write("Task %d (%s) - Index %d - %s(%d, %d)\n" % (task_ind, task_name, inst_ind, cat, a1_ind, a2_ind))
                     else:
                         raise NotImplementedError('Writing predictions is not implemented for this output_mode!')
 
@@ -464,9 +620,6 @@ def main():
                         item = test_dataset.get_labels()[task_ind][item]
                         writer.write("%s\n" % (item))
 
-    #with open(os.path.join(training_args.output_dir, 'model-summary.txt'), 'w') as model_writer:
-    #    model_writer.write(summary(model))
-
     return eval_results
 
 
